/**
  *  \file client/dialogs/selmgr.cc
  *  \brief Selection Manager Dialog
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/selmgr.h"
#include "client/widgets/selinfo.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "game/sel.h"
#include "game/univ.h"
#include "gfx/palette.h"
#include "int/interror.h"
#include "int/lex.h"
#include "int/selexpr.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "ui/dialogs/filedlg.h"
#include "game/selio.h"
#include "io/textfile.h"
#include "io/streams.h"
#include "io/dirs.h"
#include "io/glob.h"
#include "util/answer.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/quit.h"

/** Selection Manager Dialog. */
class WSelectionManager : public UIWindow {
    ObjectHolder h;

    void init();
    void updateUI();
    void doSave(string_t title, int layer);
    void doLoad(string_t title, int layer, int flags);
 public:
    WSelectionManager();
    ~WSelectionManager();

    int getCurrentLayer();
    bool handleEvent(const UIEvent& event, bool second_pass);
    void onEnter();
};

class WSelectionManagerAnswer : public AnswerProvider {
    string_t title;
 public:
    WSelectionManagerAnswer(string_t title)
        : title(title)
        { }
    bool ask(int qid, string_t question);
};

/** Constructor. */
WSelectionManager::WSelectionManager()
    : UIWindow(_("Selection Manager"), BLUE_WINDOW, COLOR_GRAY)
{
    init();
}

/** Destructor. */
WSelectionManager::~WSelectionManager()
{ }

/** Create the dialog. */
void
WSelectionManager::init()
{
    // VBox
    //   HBox
    //     VBox
    //       UIFocusGroup { WSelectionLayerInfo Layer A }
    //       ...
    //       UIFocusGroup { WSelectionLayerInfo Layer H }
    //     VBox
    //       UIButton "D-Clear"
    //       UIButton "I-Invert"
    //       UIButton "C-Copy"
    //       UIButton "S-Save"
    //       UIButton "R-Load"
    //       UIButton "M-Merge"
    //       UISpacer
    //   HBox
    //     UIButton "Enter-Select"
    //     UIButton "ESC-Close"
    //     UISpacer
    //     UIButton "H"

    setLayout(UIVBoxLayout::instance5);

    UIGroup& group1  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group11 = h.add(new UIGroup(UIVBoxLayout::instance0));
    UIGroup& group12 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group2  = h.add(new UIGroup(UIHBoxLayout::instance5));

    for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i) {
        UIFocusGroup& fg = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 1000 + i));
        fg.add(h.add(new WSelectionLayerInfo(2000 + i, i)));
        group11.add(fg);
        if (i == GMultiSelection::getCurrentSelectionLayer())
            fg.setFocusedWidget();
    }
    group1.add(group11);

    group12.add(h.add(new UIButton(_("D - Clear"),  'd')).addKey().setState(UIButton::bf_LeftJust, true));
    group12.add(h.add(new UIButton(_("I - Invert"), 'i')).addKey().setState(UIButton::bf_LeftJust, true));
    group12.add(h.add(new UIButton(_("C - Copy"),   'c')).addKey().setState(UIButton::bf_LeftJust, true));
    group12.add(h.add(new UIButton(_("S - Save"),   's')).addKey().setState(UIButton::bf_LeftJust, true));
    group12.add(h.add(new UIButton(_("L - Load"),   'l')).addKey().setState(UIButton::bf_LeftJust, true));
    group12.add(h.add(new UIButton(_("M - Merge"),  'm')).addKey().setState(UIButton::bf_LeftJust, true));
    group12.add(h.add(new UISpacer()));
    group1.add(group12);
    add(group1);

    group2.add(h.add(new UIButton(cm_Enter, _("Enter - Select"), SDLK_RETURN)).addCall(this, &WSelectionManager::onEnter));
    group2.add(h.add(new UIButton(cm_Escape, _("ESC - Close"), SDLK_ESCAPE)).addStop());
    group2.add(h.add(new UISpacer()));
    group2.add(h.add(new UIButton("H", 'h')));
    add(group2);

    add(h.add(new UIFocusIterator(1000, 1000 + GMultiSelection::NUM_SELECTION_LAYERS-1,
                                  UIFocusIterator::fi_Vert | UIFocusIterator::fi_Wrap | UIFocusIterator::fi_Tab)));
    add(h.add(new WHelpWidget("pcc2:selectionmgr")));
    add(h.add(new UIQuit(cm_Escape)));

    pack();
}

/** Update display after possible changes to the universe. */
void
WSelectionManager::updateUI()
{
    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
}

/** Save selection.
    \param title Title to use for dialogs
    \param layer Layer to save (-1 to save all) */
void
WSelectionManager::doSave(string_t title, int layer)
{
    UIFileChooserDialog select(title,
                               Directory::openDir(Directory::getWorkingDirectory()),
                               FileNamePattern::getAllFilesWithExtensionPattern("sel"));
    ui_root->centerWidget(select);
    if (ui_root->runModal(select)) {
        string_t name = appendFileNameExtension(select.getResultFile(), "sel", false);
        try {
            Ptr<Stream> file = Stream::openFile(name, Stream::C_CREATE);
            TextFile tf(*file);
            saveSelection(tf, layer, 0);
            tf.sync();
        }
        catch (FileProblemException& e) {
            messageBox(format(_("Unable to save %s: %s"), e.getFileName(), e.what()), title);
        }
    }
}

/** Load selection.
    \param title Title to use for dialogs
    \param layer Current layer (-1 not permitted here)
    \param flags Flags for loadSelection, see there. */
void
WSelectionManager::doLoad(string_t title, int layer, int flags)
{
    UIFileChooserDialog select(title,
                               Directory::openDir(Directory::getWorkingDirectory()),
                               FileNamePattern::getAllFilesWithExtensionPattern("sel"));
    ui_root->centerWidget(select);
    if (ui_root->runModal(select)) {
        string_t name = appendFileNameExtension(select.getResultFile(), "sel", false);
        try {
            WSelectionManagerAnswer answer(title);
            Ptr<Stream> file = Stream::openFile(name, Stream::C_READ);
            TextFile tf(*file);
            loadSelection(tf, layer, flags, answer);
            tf.sync();
        }
        catch (FileProblemException& e) {
            messageBox(format(_("Unable to load %s: %s"), e.getFileName(), e.what()), title);
        }
    }
}

/** Get currently focused selection layer. */
int
WSelectionManager::getCurrentLayer()
{
    // Locate a widget that has an Id in [1000, 1000+NUM_SELECTION_LAYERS).
    // This will be the selected focus group containing the current layer.
    UIBaseWidget* w = getFocusedWidget();
    while (w != 0 && w != this) {
        if (w->getId() >= 1000 && w->getId() < 1000 + GMultiSelection::NUM_SELECTION_LAYERS)
            return w->getId() - 1000;
        w = w->getParent();
    }
    return 0;
}

bool
WSelectionManager::handleEvent(const UIEvent& event, bool second_pass)
{
    static const char null_code[] = { GMultiSelection::opZero, 0 };
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'c':
         case SDLK_INSERT:
            // Copy
            {
                string_t expr;
                string_t compiledExpr;
                if (UIInputLine::run(expr, 2000, _("Selection Manager"), _("Enter layer/expression to copy from:"), 40, 0)) {
                    try {
                        IntTokenizer tok(expr);
                        compileSelectionExpression(tok, compiledExpr);
                        if (tok.getCurrentToken() != tok.tEnd)
                            throw IntError::garbageAtEnd(true);
                        GMultiSelection::executeSelectionExpression(getCurrentLayer(), compiledExpr);
                        updateUI();
                    }
                    catch (IntError& e) {
                        messageBox(format(_("Invalid selection expression: %s"), e.what()), _("Selection Manager"));
                    }
                }
            }
            return true;
         case 'd':
         case SDLK_DELETE:
            // Delete
            if (UIMessageBox(_("Do you want to clear this selection layer?"), _("Selection Manager")).doYesNoDialog()) {
                GMultiSelection::executeSelectionExpression(getCurrentLayer(), null_code);
                updateUI();
            }
            return true;
         case 'd' + ss_Ctrl:
         case SDLK_DELETE + ss_Ctrl:
            // Delete all
            if (UIMessageBox(_("Do you want to clear all selection layers?"), _("Selection Manager")).doYesNoDialog()) {
                for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i)
                    GMultiSelection::executeSelectionExpression(i, null_code);
                updateUI();
            }
            return true;
         case 'i':
         case '*':
            // Invert
            {
                int c = getCurrentLayer();
                char code[] = { GMultiSelection::opFirstLayer + c, GMultiSelection::opNot, 0 };
                GMultiSelection::executeSelectionExpression(c, code);
                updateUI();
            }
            return true;
         case 'i' + ss_Ctrl:
         case '*' + ss_Ctrl:
            // Invert all
            for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i) {
                char code[] = { GMultiSelection::opFirstLayer + i, GMultiSelection::opNot, 0 };
                GMultiSelection::executeSelectionExpression(i, code);
            }
            updateUI();
            return true;
         case 's':
            // Save
            doSave(_("Save Selection"), getCurrentLayer());
            return true;
         case 's' + ss_Ctrl:
            // Save all
            doSave(_("Save All Selections"), -1);
            updateUI();
            return true;
         case 'l':
         case 'l' + ss_Ctrl:
            // Load
            doLoad(_("Load Selection"), getCurrentLayer(), 0);
            updateUI();
            return true;
         case 'm':
         case 'm' + ss_Ctrl:
            // Merge
            doLoad(_("Merge Selection"), getCurrentLayer(), selio_Merge);
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
WSelectionManager::onEnter()
{
    GMultiSelection::setCurrentSelectionLayer(getCurrentLayer());
    stop(0);
}

bool
WSelectionManagerAnswer::ask(int /*qid*/, string_t question)
{
    return UIMessageBox(format(_("%s. Do you want to load this file?"), question), title).doYesNoDialog();
}

/** Display the "Selection Manager" dialog.
    This dialog allows the user to manipulate the current selections.
    It is invoked by the command 'UI.SelectionManager'. */
void
doSelectionManager()
{
    // Update multi-selection view from game
    GMultiSelection::getSelectionLayer(GMultiSelection::getCurrentSelectionLayer())
        .copyFrom(getDisplayedTurn().getCurrentUniverse());

    // Do dialog
    WSelectionManager mgr;
    ui_root->centerWidget(mgr);
    ui_root->runModal(mgr);
}
