/**
  *  \file int/if/ionif.cc
  *  \brief Interpreter Game Interface: Ion Storms
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/ionif.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/ion.h"
#include "game/tables.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"
#include "int/args.h"
#include "int/contextio.h"
#include "int/if/objif.h"
#include "int/propenum.h"

/* @q Storm(id:Int):Obj (Function, Context)
   Access ion storm properties.
   Use as
   | ForEach Storm Do ...
   or
   | With Storm(n) Do ...

   @diff This function was available for use in %With under the name %Storms() since PCC 1.0.11.
   Do not use the name %Storms in new code, it is not supported by PCC2; use %Storm instead.

   @see int:index:group:stormproperty|Storm Properties, int:index:group:stormcommand|Storm Commands
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFIonGet(IntArgBlock& args)
{
    int32_t iid;
    args.checkArgumentCount(1);
    if (!checkIntArg(iid, args.getNext(), 1, NUM_STORMS))
        return 0;
    return new IntIonContext(iid);
}

IntContext*
IFIonMake()
{
    IntIonContext* ic = new IntIonContext(0);
    return ic->initFirst();
}

int32_t
IFIonDim(int32_t dim)
{
    return dim == 0 ? 1 : NUM_STORMS+1;
}

/***************************** IntIonContext *****************************/

namespace {
    enum { IonProperty, IonMethod };
}

static const TIntMapping ion_mapping[] = {
    { "CLASS",     iipClass,       IonProperty, thInt },
    { "HEADING",   iipHeadingName, IonProperty, thString },
    { "HEADING$",  iipHeadingInt,  IonProperty, thInt },
    { "ID",        iipId,          IonProperty, thInt },
    { "LOC.X",     iipLocX,        IonProperty, thInt },
    { "LOC.Y",     iipLocY,        IonProperty, thInt },
    { "MARK",      0,              IonMethod,   thProcedure },
    { "MARKED",    iipMarked,      IonProperty, thBool },
    { "NAME",      iipName,        IonProperty, thString },
    { "RADIUS",    iipRadius,      IonProperty, thInt },
    { "SPEED",     iipSpeedName,   IonProperty, thString },
    { "SPEED$",    iipSpeedInt,    IonProperty, thInt },
    { "STATUS",    iipStatusName,  IonProperty, thString },
    { "STATUS$",   iipStatusFlag,  IonProperty, thBool },
    { "UNMARK",    1,              IonMethod,   thProcedure },
    { "VOLTAGE",   iipVoltage,     IonProperty, thInt },
};

static const IntObjectProcedureValue::call_t ion_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
};


/** Constructor.
    \param iid Ion Storm Id */
IntIonContext::IntIonContext(int iid)
    : IntObjectContext(current_ion_type, iid)
{ }

/** Destructor. */
IntIonContext::~IntIonContext()
{ }

int32_t
IntIonContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, ion_mapping, countof(ion_mapping));
}

void
IntIonContext::set(int32_t index, IntValue* value)
{
    switch (ion_mapping[index].domain) {
     case IonProperty:
        if (GIonStorm* ii = dynamic_cast<GIonStorm*>(getObject())) {
            setIonProperty(*ii, IntIonProperty(ion_mapping[index].index), value);
        } else {
            throw IntError::notAssignable();
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntIonContext::get(int32_t index)
{
    if (GIonStorm* ii = dynamic_cast<GIonStorm*>(getObject())) {
        switch (ion_mapping[index].domain) {
         case IonProperty:
            return getIonProperty(*ii, IntIonProperty(ion_mapping[index].index));
         case IonMethod:
            return new IntObjectProcedureValue(*ii, *getCurrentUniverse(), ion_methods[ion_mapping[index].index]);
         default:
            return 0;
        }
    } else {
        return 0;
    }
}

IntContext*
IntIonContext::clone()
{
    return new IntIonContext(getCurrentIndex());
}

string_t
IntIonContext::toString(bool /*readable*/)
{
    return format("Storm(%d)", getCurrentIndex());
}

void
IntIonContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Ion;
    sv.value = getCurrentIndex();
}

void
IntIonContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(ion_mapping, countof(ion_mapping));
}


/** Get Ion Storm property.
    \param ion Storm to query
    \param iip Property to query
    \return property value */
IntValue*
getIonProperty(const GIonStorm& ion, IntIonProperty iip)
{
    if (ion.getVoltage() == 0)
        return 0;
    switch (iip) {
     case iipClass:
        /* @q Class:Int (Storm Property)
           Ion storm's class. */
        return makeIntValue(ion.getClass());
     case iipHeadingInt:
        /* @q Heading$:Int (Storm Property)
           Ion storm's heading, in degrees. */
        return makeIntValue(ion.getHeading());
     case iipHeadingName:
        /* @q Heading:Str (Storm Property)
           Ion storm's heading, as compass point.
           For example, "NE" for north-east. */
        return makeStringValue(getHeadingName(ion.getHeading()));
     case iipId:
        /* @q Id:Int (Storm Property)
           Ion storm's Id. */
        return makeIntValue(ion.getId());
     case iipLocX:
        /* @q Loc.X:Int (Storm Property)
           Ion storm center X coordinate. */
        return makeIntValue(ion.getPos().x);
     case iipLocY:
        /* @q Loc.Y:Int (Storm Property)
           Ion storm center Y coordinate. */
        return makeIntValue(ion.getPos().y);
     case iipMarked:
        /* @q Marked:Bool (Storm Property)
           True if ion storm is marked. */
        return makeBoolValue(ion.isMarked());
     case iipName:
        /* @q Name:Str (Storm Property)
           Ion storm name. */
        return makeStringValue(ion.getName(GObject::PlainName));
     case iipRadius:
        /* @q Radius:Int (Storm Property)
           Ion storm radius in ly. */
        return makeIntValue(ion.getRadius());
     case iipSpeedInt:
        /* @q Speed$:Int (Storm Property)
           Ion storm speed (warp factor). */
        return makeIntValue(ion.getSpeed());
     case iipSpeedName:
        /* @q Speed:Str (Storm Property)
           Ion storm speed, as human-readable string. */
        return makeStringValue(format("Warp %d", ion.getSpeed()));
     case iipStatusFlag:
        /* @q Status$:Bool (Storm Property)
           Ion storm status.
           - True if storm is growing
           - False if storm is weakening */
        return makeBoolValue(ion.isGrowing());
     case iipStatusName:
        /* @q Status:Str (Storm Property)
           Ion storm status, as human-readable string. */
        return makeStringValue(ion.isGrowing() ? "Growing" : "Weakening");
     case iipVoltage:
        /* @q Voltage:Int (Storm Property)
           Ion storm voltage, in MeV. */
        return makeIntValue(ion.getVoltage());
    }
    return 0;
}

/** Set Ion Storm property. */
void
setIonProperty(GIonStorm& /*ion*/, IntIonProperty /*iip*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}
