/**
  *  \file int/process.cc
  *  \brief Interpreter: Process Management
  *
  *  We distinguish between two kinds of processes, regular processes
  *  and temporary processes.
  *
  *  Regular processes are added to the process list normally, and can
  *  do everything they might want to do, including suspending. Such
  *  processes are created using createProcess. The user then sets up
  *  the process and runs it using runRunnableProcesses,
  *  killTerminatedProcesses
  *
  *  Temporary processes are used whenever the PCC2 core needs to
  *  evaluate an expression, and wishes immediate feedback. Such
  *  processes are not allowed to suspend; when they try to, they
  *  receive a (catchable) error. Such processes are created by simply
  *  creating a IntExecutionContext object on the stack, setting it
  *  up, and and giving it to runTemporaryProcess. If that returns
  *  true, the process has produced its result as requested, otherwise
  *  it didn't. The process never appears in the process list, and is
  *  killed automatically when the IntExecutionContext object goes out
  *  of scope. This is the equivalent to PCC 1.x's simple process-less
  *  expression evaluation. PCC2 doesn't support process-less
  *  evaluation because it supports user-defined functions.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/process.h"
#include "int/exec.h"
#include "util/misc.h"
#include "game/obj.h"

/******************************* Public API ******************************/

typedef ptr_vector<IntExecutionContext> ProcessList;

static IntProcessList process_list;

/** Create a regular process. The process is created in state psRunnable.
    You have to fill it (by pushing some frames), and then run it by
    runRunnableProcesses().
    \param name Name of process
    \return Reference to blank process object */
IntExecutionContext&
createProcess(string_t name)
{
    IntExecutionContext* exc = new IntExecutionContext(name);
    process_list.push_back(exc);
    handlePriorityChange(*exc);
    return *exc;
}

/** Run runnable processes. Runs all processes that are ready to run
    (psRunnable). */
void
runRunnableProcesses()
{
    // Must use indexes instead of iterators here, because running a process can create
    // new processes and thus change the process list!
    for (uint32_t i = 0; i != process_list.size(); ++i) {
        IntExecutionContext* exc = process_list[i];
        if (exc->getState() == IntExecutionContext::psRunnable) {
            exc->run();
            if (exc->getState() == IntExecutionContext::psRunnable) {
                // If the process is still runnable, it must be because it performed
                // some UI operation. Exit, and have the caller execute the required
                // operation and call us again.
                break;
            }
        }
    }
}

/** Terminate all processes. */
void
terminateAllProcesses()
{
    for (IntProcessList::iterator i = process_list.begin(); i != process_list.end(); ++i) {
        if ((*i)->getState() != IntExecutionContext::psFrozen) {
            (*i)->setState(IntExecutionContext::psTerminated);
        }
    }
}

/** Kill terminated processes. Terminates all processes that have exited
    (psTerminated, psError, psEnded). */
void
killTerminatedProcesses()
{
    IntProcessList::size_type i = 0, o = 0;
    while (i < process_list.size()) {
        switch (process_list[i]->getState()) {
         case IntExecutionContext::psTerminated:
         case IntExecutionContext::psTerminating:
         case IntExecutionContext::psError:
         case IntExecutionContext::psEnded:
            // Killing a process invoked from an object marks that object dirty,
            // in order to remove the "there's a process here" marker.
            if (GObject* obj = process_list[i]->getInvokingObject()) {
                obj->markDirty();
            }
            ++i;
            break;
         default:
            process_list.swap_elems(i, o);
            ++i;
            ++o;
            break;
        }
    }
    while (o < process_list.size()) {
        process_list.pop_back();
    }
}

/** Check whether there's any runnable process. */
bool
hasAnyRunnableProcess()
{
    for (IntProcessList::iterator i = process_list.begin(); i != process_list.end(); ++i) {
        if ((*i)->getState() == IntExecutionContext::psRunnable) {
            return true;
        }
    }
    return false;
}

/** Run temporary process. Runs the given process, which is a stack
    object of the caller (not created with createProcess). The process
    is not permitted to suspend.

    \retval true  The process has run till its end and finished normally.
                  If it contained an expression, that expression's result
                  is available on the stack (exec.getResult()).
    \retval false The process has terminated abnormally. This happens when
                  the process received an error, but also when it
                  executed an 'End' statement. In this case, it has not
                  produced a result. */
bool
runTemporaryProcess(IntExecutionContext& exec)
{
    while (1) {
        exec.run();

        /* Do not terminate */
        if (exec.getState() == IntExecutionContext::psSuspended) {
            exec.handleException("Cannot suspend temporary process", string_t());
        }

        /* Process status */
        if (exec.getState() == IntExecutionContext::psRunnable) {
            /* It wants to run again */
        } else if (exec.getState() == IntExecutionContext::psEnded) {
            return true;
        } else {
            // FIXME what to do with UI-wait processes?
            // psError, psTerminated, ...
            return false;
        }
    }
}

/** Handle priority change. If this causes the process's location in
    the process list to change, move it accordingly. */
void
handlePriorityChange(IntExecutionContext& proc)
{
    /* FIXME: check the interaction of this routine and runRunnableProcesses.
       This will change the list runRunnableProcesses is iterating. */

    /* Locate the process. Search backward, because a common case is
       that the last process changes its priority, as a result of
       being loaded or created. */
    uint32_t pos = process_list.size();
    while (pos > 0 && &proc != process_list[pos-1]) {
        --pos;
    }
    if (pos == 0) {
        /* can't happen */
        return;
    }
    --pos;

    /* Sanity check: */
    if (pos > 0 && proc.getPriority() < process_list[pos-1]->getPriority()) {
        /* Low value (high priority), move process to front */
        do {
            process_list.swap_elems(pos, pos-1);
            --pos;
        } while (pos > 0 && proc.getPriority() < process_list[pos-1]->getPriority());
    }
    if (pos < process_list.size()-1 && proc.getPriority() > process_list[pos+1]->getPriority()) {
        /* High value (low priority), move process to back.
           Note ">" vs. ">=" in the conditions: we move the process
           only if it's out of place, but when we move it, we move it
           to the end of its priority. */
        do {
            process_list.swap_elems(pos, pos+1);
            ++pos;
        } while (pos < process_list.size()-1 && proc.getPriority() >= process_list[pos+1]->getPriority());
    }
}

/** Get process given an object.
    \param obj Invoking object
    \param kind Process kind
    \return process if any, 0 if none */
IntExecutionContext*
getProcessByObject(const GObject* obj, uint8_t kind)
{
    for (IntProcessList::iterator i = process_list.begin(); i != process_list.end(); ++i) {
        if ((*i)->getProcessKind() == kind && (*i)->getInvokingObject() == obj) {
            return *i;
        }
    }
    return 0;
}

/** Get process list. Note that you may modify the processes, but not
    the list. */
const IntProcessList&
getProcessList()
{
    return process_list;
}
